#!/usr/bin/perl
# $Id: namistai.pl,v 1.49 2004/04/17 11:46:53 michiel Exp $
#
# namistai.pl - a http2smb proxy.
#
# uses Filesys::SmbClient for the smb interaction,
# HTTP::Daemon and HTTP::Status for the HTTP handling.
# reloads mime_types and configfile on HUP,
# clean exit on TERM.
# expects an index.html in it's cwd.
#
# mad idea by Michiel Holtkamp, code by Michiel Holtkamp & Gertjan Haaijer.
# contact us at :
# michiel@namistai.nl.eu.org
# gertjan@namistai.nl.eu.org

use strict;
use warnings;
use POSIX;
use Filesys::SmbClient;
use Getopt::Std;
use HTTP::Daemon;
use HTTP::Status;
use sigtrap 'handler', \&sighandler, 'normal-signals';

# i have yet to find out what these do, i only know that they work.
use constant RES_HEADERS => qw(Pragma no-cache Connection close);

use constant PATH_SIZE_DL => "\0";

# function prototypes

sub mime_read($);
sub config_read($);
sub get_mime_type($);
sub get_dir_list($);
sub smb_get_dir_list($);
sub transfer_file($$$);
sub format_dir_list($);
sub create_http_response($$$$);
sub sighandler($);
sub child_httpd($);

# getopt options, have to be declared "our".
our ($opt_p, $opt_i, $opt_h, $opt_c, $opt_m);

# declare object $smb here, but create it after configfile has been read.
my $smb;

# various global variabeles and default values.

my $DELIMITER = ";";
my %acl_hosts_accept = ();
my @mime_types;
my $sig = 0;
my $port = 2000;
my $host = "127.0.0.1";
my $smb_user = "guest";
my $smb_pass = "guest";
my $wget = 0;

sub main()
{
	my $argc = @ARGV;
	my ($httpd, $t, $c, $r, $pid, $tmp);
	my @list;
	my $config = "namistai.cfg";
	my $mime = "/etc/mime.types";
	
	# place our sighandler so we don't end up with 4373435 zombies,
	# and so we can rehash configfile with a HUP signal.

	$SIG{"CHLD"} = \&sighandler;
	$SIG{"HUP"} = \&sighandler;
	$SIG{"TERM"} = \&sighandler;

	print "namistai - a http2smb proxy\n";

	getopts("m:i:p:c:h");
	
	if (defined($opt_h)) {
		print "options: [-i host/ip] [-p port] [-c configfile]" .
			"[-m mime_types_file]\n";
		exit(1);
	}

	print "use -h for help\n";

	if (defined($opt_c)) {
		$config = $opt_c;
	}

	# read config first, other settings take precedence over config
	config_read($config);

	# create smb object, this will handle all the smb connections.
	$smb = new Filesys::SmbClient(username => $smb_user,
									password => $smb_pass,
									workgroup => "",
									debug => 0);

	if (!defined($smb)) {
		die "Can't make Filesys::SmbClient, $!\n";
	}
	
	if (defined($opt_p)) {
		$port = $opt_p;
	}
	if (defined($opt_i)) {
		$host = $opt_i;
	}
	if (defined($opt_m)) {
		$mime = $opt_m;
	}

	if (!-e "index.html") {
		print "warning, no index.html found!\n";
	}
	
	# get the available mime types on this system.
	mime_read($mime);

	$httpd = new HTTP::Daemon->new(LocalAddr => $host,
									LocalPort => $port);
	
	if (!defined($httpd)) {
		die "Can't make HTTP::Daemon, $!\n";
	}
	
	print "daemonizing\n";
	$pid = fork();

	if (defined($pid) && $pid == 0) {
		while (1) {
			while ($c = $httpd->accept) {
				$pid = fork();
				if (defined($pid) && $pid == 0) {
					child_httpd($c);
					exit(1);
				}
			}
			
			if ($sig == 1) {
				undef(%acl_hosts_accept);
				undef(@mime_types);
				config_read($config);
				mime_read($mime);
				$sig = 0;
			} elsif ($sig == 2) {
				print "closing sockets: ";
				if ($c) {
					$c->close;
				}
				$httpd->close;
				print ".\n";
				die "received SIGTERM\n";
			}
		}
	}

	return;
}

sub child_httpd($)
# this is a forked child. it handles HTTP requests.		
{
	my $c = shift;
	my ($t, $tmp, $r);
		
	if (!defined($acl_hosts_accept{$c->peerhost()})) {
		next;
	}

	while (defined($t = $c->get_request)) {
		if ($t->header('user_agent') =~ /Wget/) {
			$wget = 1;
		} else {
			$wget = 0;
		}

		if ($t->method eq "GET" && ($t->url->path eq "/" || 
			 $t->url->path eq "/index.html"))
		{
			if (-e "index.html") {	
				$c->send_file_response("index.html");
			} else {
				$c->send_error(404, "index.html not found!");
			}
		} elsif ($t->method eq "GET" && $t->url->path =~ /\/$/) 
		{
			# get the smb dirlisting 
			
			$tmp = get_dir_list($t->url->path);
			if (!$tmp) {
				$c->send_error(404, "couldn't get dirlisting : $!");
			} else {
				# put it in a http response message
				$r = create_http_response($tmp, "text/html", "", "");
				if (!defined($r)) {
					$c->send_error(404, "no http response: $!");
				}
	
				# send the http response message
				$c->send_response($r);
			}
		} else {
			my $range;

			$range = $t->header('Range');
			
			if (!defined($range)) {
				$range = "-";
			}
	
			# now, get the file piece by piece and send them
			# directly to the browser.
			transfer_file($t->url->path, $c, $range);
		}
		# next, close the socket, so the browsers knows we're done.
		$c->close;

		# don't forget to set $wget to 0,
		# so we won't end up always being in "wget mode".
		$wget = 0;
	}
	return;
}

sub create_http_response($$$$)
# creates a HTTP::Response object, with given content and type.
# if no type is given, "application/octet-stream" is assumed.
{
	my $r;
	my $content = shift;
	my $type = shift;
	my $length = shift;
	my $range = shift;

	$r = HTTP::Response->new(200, "Ok", new HTTP::Headers(RES_HEADERS));
	if (!defined($r)) {
		return;
	}

	$r->protocol("HTTP/1.1");
	
	if (!defined($type) || $type eq "") {
		$r->header('Content-type'=>'application/octet-stream');
	} else {
		$r->header('Content-type'=>$type);
	}

	if (defined($length) && $length ne "") {
		$r->header('Content-Length'=>$length);
	}

	if (defined($range) && $range ne "") {
		$r->header('Content-Range'=>$range);
	}

	$r->content($content);

	return($r);
}

sub get_mime_type($)
# parses an array containing mime_types. returns the type of
# content if a matching extention is found.
{
	my $url = shift;
	my @entry;
	my $type;
	my @exts;
	
	# get the ext
	$url =~ s/.*\.([^.]*)/$1/;
		
	foreach (@mime_types) {
		@entry = split(/:/, $_, 2);
		@exts = split(/ /, $entry[1]);
		foreach (@exts) {
			if ($url eq $_) {
				return $entry[0];
			}
		}
	}

	return;
}

sub get_dir_list($)
# get dir listing from share, then convert it to html.
# return a string containing <a href> style links.
{
	my $share = shift;
	my $t;
	my @list;

	@list = smb_get_dir_list($share);
	if (!@list) {
		return;
	}

	$t = join($DELIMITER, @list);
	@list = format_dir_list($t);
	$t = join(" ", @list);

	return($t);
}

sub smb_get_dir_list($)
# get dir listing from a share,
# directory entries will be appended with a '/'.
# list gets sorted alphabetically.
{
	my $share = shift;
	my $fd;
	my @list;
	my @stat;

	if ($share !~ /^smb:\//) {
		$share = "smb:/" . $share;
	}

	$share =~ s/\/$//;
	$share =~ s/\'/%27/;

	$fd = $smb->opendir($share);
	if (!defined($fd) || !$fd) {
		return;
	}

	# append a slash incase it's not there,
	# otherwise we get weird paths.
	if($share !~ /\/$/) {
		$share .= "/";
	}
	
	foreach ($smb->readdir_struct($fd)) {
		if ($_->[0] == SMBC_DIR ||
				$_->[0] == SMBC_FILE_SHARE) {
			if ($_->[1] eq "." || $_->[1] eq "..") {
				next;
			}
			push(@list, $share . $_->[1] . "/");
		} elsif ($_->[0] == SMBC_FILE) {
			@stat = $smb->stat($share . $_->[1]);
			if ($#stat == 0) {
				next;
			}
			push(@list, $share . $_->[1] . PATH_SIZE_DL . $stat[7]);
		}
	}

	$smb->close($fd);

	@list = sort(@list);
	unshift(@list, $share . "../");
	unshift(@list, $share . "./");
	
	return(@list);
}

sub format_dir_list($)
# convert a dir listing (handed in "<e>;<e>" format) to html.	
{
	my $tmp = shift;
	my @list;
	my @entry;
	my @newlist;
	my (@left, @right);
	my ($i, $t, $size, $total_size);
	my $del = PATH_SIZE_DL;
	
	@list = split(/$DELIMITER/, $tmp);
	
	push(@newlist, "<html>");
	push(@newlist, "<body>");
	push(@newlist, "<table height=\"100%\" width=\"100%\" border=0 cellspacing=0 cellpadding=0>");

	foreach (@list) {
		# split our smb link in parts, so we can extract the file/
		# dirname to avoid very loooooong paths in our html.
		
		@entry = split(/\//, $_);
		$t = $entry[$#entry];

		# append / if we're a directory
		if ($_ =~ /\/$/) {
			$t .= "/";
			$i = "<a href=\"$_\">$t</a><br>\n";
		} else {
			@entry = split(/$del/, $t);
			$t = $entry[0];
			chomp($t);
			$size = $entry[1];
			@entry = split(/$del/, $_);
			$_ = $entry[0];
			$i = "<a href=\"$_\">$t</a>";
			$_ =~ s/\///;
			if (!$wget) {
				$i .= " <a href=\"/force_text.pl=$_\">[T]</a><br>\n";
			} else {
				$i .= "<br>\n";
			}
		}

		$i =~ s/smb:\///g;

		push(@left, $i);
		if (defined($size) && $size != 0) {
			$total_size += $size;
			push(@right, $size . "<br>");
			undef $size;
		} else {
			push(@right, "<br>");
		}
	}

	push(@newlist, "<tr>");
	push(@newlist, "<td width=200 valign=top>");

	foreach (@left) {
		push(@newlist, $_);
	}
	
	push(@newlist, "</td>");
	push(@newlist, "<td width=200 align=right valign=top>");
	
	foreach (@right) {
		push(@newlist, $_);
	}
	
	if (defined($total_size) && $total_size != 0) {
		$i = "Total size: ";
		if ($total_size < 1024) {
			$i .= "$total_size ";
		} else {
			$total_size /= 1024;
			if ($total_size < 1024) {
				$i .= sprintf("%.0fK\n", $total_size);
			} else {
				$total_size /= 1024;
				if ($total_size < 1024) {
					$i .= sprintf("%.0fM\n", $total_size);
				} else {
					$total_size /= 1024;
					if ($total_size < 1024) {
						$i .= sprintf("%.0fG\n", $total_size);
					} else {
						$total_size /= 1024;
						$i .= sprintf("%.0fT\n", $total_size);
					}
				}
			}
		}

		$i .= "<br>\n";
		push(@newlist, $i);
	}

	push(@newlist, "</td>");
	push(@newlist, "</tr>");
	push(@newlist, "</table>");
	push(@newlist, "</body>");
	push(@newlist, "</html>");

	return(@newlist);
}

sub transfer_file($$$)
# opens a shared file, then reads and sends the contents
# of that file, 4k at a time, to the socket given as 2nd parameter.
{
	my $t = shift;
	my $share = $t;
	my ($fd, $c, $buf);
	my $socket = shift;
	my $range = shift;
	my $range_begin;
	my $range_end;
	my @stat;
	my $r;
	my $sendbytes;
	my $force_text = 0;
	my $type;

	if ($share !~ /smb:\//) {
		$share = "smb:/" . $share;
	}

	if ($share =~ /^smb:\/\/force_text\.pl=/) {
		$force_text = 1;
		$share =~ s/force_text\.pl=//;
	}
	
	@stat = $smb->stat($share);
	if ($#stat == 0) {
		$socket->send_error(404, "smb stat failed: $!");	
		return;
	}

	$range =~ s/bytes=//;
	($range_begin, $range_end) = split(/-/, $range);
	$range = "bytes=" . $range;
	if ($range_begin eq "") {
		$range_begin = 0;
	}
	if ($range_end eq "") {
		$range_end = $stat[7] - 1;
	}
	
	$fd = $smb->open($share);

	if (!defined($fd) || $fd == 0) {
		$socket->send_error(404, "smb stat failed: $!");	
		return;
	}

	$c = $smb->seek($fd, $range_begin);
	if ($c < 0) {
		$socket->send_error(404, "seek failed: $!");
		return;
	}

	# first, send a response with empty content, 
	# cos we want to handle our own file sending stuff.
	# the response depends on the mime type,
	# we do want to be able to view certain files directly.
 
	if ($force_text == 0) {
		$type = get_mime_type($t);
	} else {
		$type = "text/plain";
	}

	$r = create_http_response("" , $type, $range_end - $range_begin + 1,
			"bytes $range_begin-$range_end/$stat[7]");
	$socket->send_response($r);

	# read data from an open smb shared file,
	# and push it directly to teh browser.
	# XXX this isn't very nice, because we ignore the $range_end parameter.
	while (defined($c = $smb->read($fd))) {
		$sendbytes = send($socket, $c, 0);
		if (!defined($sendbytes)) {
			last;
		}
	}

	$smb->close($share);
	return;
}

sub config_read($)
# reads the configfile
#
{
	my $config = shift;

	unless (open(CONFIG, '<', $config)) {
		print STDERR "Can't open $config: $!\n";
		return;
	}

	while (<CONFIG>) {
		/^#/ && next;	# skip comments

		if (/accept_host=/) {
			$_ =~ s/accept_host=//;
			chomp;
			foreach (split(/,/)) {
				$acl_hosts_accept{$_} = 1;
			}
			next;
		}
		/port=/ && do { $_ =~ s/port=//; chomp; $port = $_; next; };
		/host=/ && do { $_ =~ s/host=//; chomp; $host = $_; next; };
		/user=/ && do { $_ =~ s/user=//; chomp; $smb_user = $_; next; };
		/pass=/ && do { $_ =~ s/pass=//; chomp; $smb_pass = $_; next; };
		chomp;
		print("unknown option: $_\n");
	}

	close CONFIG;
}

sub mime_read($)
# parses a mime.types file
# stores in global array mime_types.
{
	my $file = shift;
	my @entry;
	my $i;

	if (@mime_types) {
		undef(@mime_types);
	}

	unless (open(INPUT, $file)) {
		print STDERR "Can't open $file: $!\n";
		return;
	}

	while (<INPUT>) {
		if ($_ !~ /^#/) {	
			chomp($_);
			@entry = split(/\s/, $_, 2);
			if ($entry[0] && $entry[1]) {
				$entry[1] =~ s/^\s*//g;
				$i = $entry[0] . ":" . $entry[1];
				push(@mime_types, $i);
			}
		}
	}
	close INPUT;
	return;
}

sub sighandler($)
# trap signal, (set int | wait), return.	
{
	my $i = shift;
	
	for ($i) {
		/HUP/ && do { $sig = 1; };
		/TERM/ && do { $sig = 2; };
		/CHLD/ && do { wait; }; 
	}

	$SIG{$i} = \&sighandler;
}

&main ();
